
import qgis
import qgis.core

# abstract class Waterbody is the root class of our hierarchy 
class Waterbody():
    
    # constructor (can be derived by subclasses)
    def __init__(self, name, geometry):
        self.name = name            # instance variable for storing the name of the watebrbody
        self.geometry = geometry    # instance variable for storing the a QgsGeometry object with the geometry for this waterbody

    # abstract static class function for creating a waterbody object if the given way satisfies
    # the required conditions; needs to be overridden by instantiable subclasses 
    def fromOSMWay(way, allNodes):     
        pass
    
    # abstract method for creating QgsFeature object for this waterbody;
    # needs to be overridden by instantiable subclasses 
    def toQgsFeature(self):
        pass
    

# abstract class LinearWaterBody is derived from class Waterbody
class LinearWaterbody(Waterbody):
    
    # constructor (can be invoked by derived classes and takes care of the length computation)
    def __init__(self, name, geometry):
        super(LinearWaterbody, self).__init__(name, geometry)
        
        # calculate length of this linear waterbody
        qda = qgis.core.QgsDistanceArea() 
        qda.setEllipsoid('WGS84')
        length = qda.measureLength(geometry)

        # instance variable for storing the length of this linear waterbody
        self.length = qda.convertLengthMeasurement(length, qgis.core.QgsUnitTypes.DistanceMeters) 

    # ... you may want to add additional auxiliary methods or class functions to this class definition


# abstract class ArealWaterbody is derived from class Waterbody
class ArealWaterbody(Waterbody):

    # constructor (can be invoked by derived classes and takes care of the area computation)
    def __init__(self, name, geometry):
        super(ArealWaterbody, self).__init__(name, geometry)

        # calculate area of this areal waterbody
        qda = qgis.core.QgsDistanceArea() 
        qda.setEllipsoid('WGS84')
        area = qda.measureArea(geometry)

        # instance variable for storing the length of this areal waterbody
        self.area = qda.convertAreaMeasurement(area, qgis.core.QgsUnitTypes.AreaSquareMeters)

    # ... you may want to add additional auxiliary methods or class functions to this class definition


# class Stream is derived from class LinearWaterBody and can be instantiated
class Stream(LinearWaterbody):
    
    # constructor (calls LinearWaterbody constructor to initialize name, geometry, and length instance variables)
    def __init__(self, name, geometry):
        super(Stream,self).__init__(name, geometry)

    # override the fromOSMWay(...) static class function
    def fromOSMWay(way, allNodes):
        # ... write code that tests whether the way element in parameter 'way' satisfies the tag conditions for Streams;
        # if yes:
        #     - get the name of the way element in 'way' if its tags dictionary contains the entry 'name'; else use the
        #       string 'unknown' for the name 
        #     - use the way's node list to create a list of QgsPointXY objects for the given way element; the 
        #       lat and lon coordinates for each node ID in the list can be looked up in the dictionary provided in 
        #       the 'allNodes' parameter of the function
        #     - create the polyline QgsGeometry object from the QgsPointXY list using qgis.core.QgsGeometry.fromPolylineXY(...) 
        #       as in the lesson walkthrough code (Section 4.10.2.4)
        #     - create the new Stream object by calling the constructor Stream(...) with the name and QgsGeometry object
        #       as paramters and return the new Stream object
        # if no: simply return None
        pass # we just put this here so that we don't get a syntax error while the function is not implemented yet

    # override the toQgsFeature(...) method
    def toQgsFeature(self):
        # ... create a new QgsFeature object for this Stream with NAME, TYPE, and LENGTH attributes and return it; as 
        # in Section 4.5.3 and the walkthrough code, you will have to create a new QgsFeature object for this and then
        # set its geometry (setGeometry(...)) and its attributes (setAttributes(...)) before you return it
        pass # we just put this here so that we don't get a syntax error while the function is not implemented yet

    # Don't forget to add a definition of the method __str__(self) that returns a string describing the object including its
    # name, type, and length, e.g. Stream Rapid Run (length: 372.01922201444535m)  


